/*******************************************************************************
 * Copyright 2012 Mohamed Sellami, Telecom SudParis
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package telecom.sudparis.eu.paas.core.server.ressources.manager.environment;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;

import javax.ws.rs.Path;
import javax.ws.rs.core.GenericEntity;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import javax.xml.transform.stream.StreamSource;

import org.cloudfoundry.client.lib.domain.Staging;

import com.amazonaws.AmazonClientException;

import telecom.sudparis.eu.paas.api.ressources.manager.environment.RestEnvironmentManager;
import telecom.sudparis.eu.paas.core.server.pool.environment.EnvironmentPool;
import telecom.sudparis.eu.paas.core.server.ressources.exception.NotSupportedException;
import telecom.sudparis.eu.paas.core.server.ressources.util.EnvironmentLinkGenerator;
import telecom.sudparis.eu.paas.core.server.xml.Error;
import telecom.sudparis.eu.paas.core.server.xml.LinksListType;
import telecom.sudparis.eu.paas.core.server.xml.OperationResponse;
import telecom.sudparis.eu.paas.core.server.xml.environment.ConfigurationType;
import telecom.sudparis.eu.paas.core.server.xml.environment.EntryType;
import telecom.sudparis.eu.paas.core.server.xml.environment.EnvironmentType;
import telecom.sudparis.eu.paas.core.server.xml.environment.list.EnvironmentsType;
import telecom.sudparis.eu.paas.core.server.xml.environment.list.SimpleEnvironmentType;
import telecom.sudparis.eu.paas.core.server.xml.manifest.PaasEnvironmentNodeType;

/**
 * REST resource of type EnvironmentManager
 * 
 * @author Mehdi Ahmed-Nacer (Telecom SudParis)
 * 
 */
@Path("environment")
public class EnvironmentManagerRessource implements RestEnvironmentManager {

	/**
	 * ressource bundle to get the connexion credentials
	 */
	private static ResourceBundle rb = ResourceBundle
			.getBundle("telecom.sudparis.eu.paas.core.server.ressources.credentials");

	/**
	 * The public CF-PaaS API URL
	 */
	private static String apiUrl = rb.getString("api.public.url");

	/**
	 * An element to display The response
	 */
	private OperationResponse or = new OperationResponse();

	/**
	 * An element to display Errors
	 */
	private Error error = new Error();

	/**
	 * The Frameworks supported in the current implementation
	 */
	public static final String JAVA_WEB = "java_web";
	public static final String SPRING_WEB = "spring_web/1.0";
	public static final String NODE = "node";

	/**
	 * The Runtimes supported in the current implementation
	 */
	public static final String JAVA6 = "java";
	public static final String JAVA7 = "java7";

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Response createEnvironment(String environmentTemplateDescriptor) {

		CloudFoundaryManageEnvironment cf = new CloudFoundaryManageEnvironment();
		EnvironmentType resCf = cf.createEnvironment(environmentTemplateDescriptor);
		
		AmazonManageEnvironment aws = new AmazonManageEnvironment();
		EnvironmentType resAws = aws.createEnvironment(environmentTemplateDescriptor);
		
		if(resCf == null || resAws == null)
			throw new NotSupportedException(resCf+","+resAws );
		
		or.setValue("EnvId CF: "+ resCf.getEnvId()+", AWS: "+resAws.getEnvId());
		//or.setValue("EnvId AWS: "+ resCf.getEnvId());
		return Response.status(Response.Status.OK)
				.entity(or).type(MediaType.APPLICATION_XML_TYPE).build();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Response deleteEnvironment(String envid) {
		try {
			boolean result = EnvironmentPool.INSTANCE.delete(envid);
			if (result) {
				or.setValue("The environment " + envid
						+ " was succesfully deleted from the Environment pool.");
				return Response.status(Response.Status.OK).entity(or)
						.type(MediaType.APPLICATION_XML_TYPE).build();
			} else {
				or.setValue("The environment " + envid
						+ " was not found in the Environment pool.");
				return Response.status(Response.Status.ACCEPTED).entity(or)
						.type(MediaType.APPLICATION_XML_TYPE).build();
			}
		} catch (Exception e) {
			System.out.println("Failed to delete the environment: "
					+ e.getMessage());
			e.printStackTrace();
			error.setValue("Failed to delete the environment: "
					+ e.getMessage());
			return Response.status(Response.Status.INTERNAL_SERVER_ERROR)
					.entity(error).type(MediaType.APPLICATION_XML_TYPE).build();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Response findEnvironments() {
		List<SimpleEnvironmentType> envList = formatEnvList(EnvironmentPool.INSTANCE
				.getEnvList());
		EnvironmentsType envs = new EnvironmentsType();
		envs.setEnvironment(envList);

		if (envList != null) {
			return Response.status(Response.Status.OK)
					.entity(new GenericEntity<EnvironmentsType>(envs) {
					}).type(MediaType.APPLICATION_XML_TYPE).build();
		} else {
			System.out.println("Failed to retrieve the environments list.");
			error.setValue("Failed to retrieve the environments list.");
			return Response.status(Response.Status.NO_CONTENT).entity(error)
					.build();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Response getEnvironment(String envid) {
		try {
			EnvironmentType env = EnvironmentPool.INSTANCE.getEnv(envid);
			if (env != null) {
				return Response.status(Response.Status.OK)
						.entity(new GenericEntity<EnvironmentType>(env) {
						}).type(MediaType.APPLICATION_XML_TYPE).build();
			} else {
				System.out.println("Failed to find the cloud Environment "
						+ envid);
				error.setValue("Failed to find the cloud Environment " + envid);
				return Response.status(Response.Status.ACCEPTED).entity(error)
						.build();
			}
		} catch (Exception e) {
			System.out.println("Failed to display the environment: "
					+ e.getMessage());
			e.printStackTrace();
			error.setValue("Failed to display the environment: "
					+ e.getMessage());
			return Response.status(Response.Status.INTERNAL_SERVER_ERROR)
					.entity(error).type(MediaType.APPLICATION_XML_TYPE).build();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Response updateEnvironment(String envid,
			String environmentTemplateDescriptor) {
		throw new NotSupportedException(
				"The updateEnvironment is not yet implemented.");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Response getDeployedApplications(String envid) {
		throw new NotSupportedException(
				"The getDeployedApplications is not yet implemented.");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Response getInformations() {
		throw new NotSupportedException(
				"The getInformations is not yet implemented.");
	}

	// private methods
	private synchronized Long getNextId() {
		return new Long(EnvironmentPool.INSTANCE.getNextID());
	}

	private List<SimpleEnvironmentType> formatEnvList(
			List<EnvironmentType> envList) {
		List<SimpleEnvironmentType> listEnvType = new ArrayList<SimpleEnvironmentType>();
		for (EnvironmentType e : envList) {
			SimpleEnvironmentType envType = new SimpleEnvironmentType();
			envType.setDescription(e.getEnvDesc());
			envType.setId(e.getEnvId());
			envType.setName(e.getEnvName());
			envType.setUri(EnvironmentLinkGenerator.formatApiURL(apiUrl)
					+ "environment/" + e.getEnvId());
			listEnvType.add(envType);

		}
		return listEnvType;
	}

}